/*  @(#)learn.mod	1.4   12/9/92   
 *
 *  File : learn.mod
 *  Simulation of primate visual/motor conditional learning.
 *
 *  Andrew H. Fagg
 *  Center for Neural Engineering
 *  Computer Science Department
 *  University of Southern California
 *  Henry Salvatori #300
 *  Los Angeles, CA  90089-0781
 *
 *  ahfagg@pollux.usc.edu
 *
 *
 *  For a description of the model, see :
 *   Fagg, Andrew H., Arbib, Michael A., "A model of primate visual/motor
 *      conditional learning", Journal of Adpative Behavior, Summer 1992,
 *      pp. 3 - 37
 *
 *  For a description of how to use this model, see the file README
 *
 *  Implementation in NSL (Neural Simulation Language), see
 *   Weitzenfeld, A., "NSL: Neural simulation language, version 2.1",
 *     Technical Report 91-05.  Lost Angeles: Center for Neural Engineering,
 *     University of Southern California.
 *
 * HISTORY
 *  Date         Author               Description
 *  --------     -----------------    --------------------------------
 *  12/05/92     Andrew H. Fagg       Doc update.
 *  01/15/92     Andrew H. Fagg       Original.
 *
 */

// Modifications :
//	
//	1) Create nslClass LearnMethods
//	2) normal_row -> LearnMethods.normal_row
//	3) normal_col -> LearnMethods.normal_col
//	4) select_random -> LearnMethods.select_random
//	5) input -> input1
//	6) output -> output1
//	7) w_in_feature_probability -> w_in_featureProbability
//	8) w_vote_motor_probability -> w_vote_motorProbability
//	9) voting_contribution_scale -> votingContributionScale
//	10) LIMITED_VOTING_ACTIVITY_FLAG -> LIMITED_VOTING_ACTIVITY
//      11) nsl_vector -> NslVector
//      12) nsl_matrix -> NslMatrix
//      13) nsl_file -> NslFile

import alib;
import learn;


nslClass LearnMethods ()
{
 void normal_row(NslMatrix& mat, int norm_mode)
 {
  int i,j;
  int a,b;
  float count;
				// Get size of matrix 
  a = mat.get_xn();
  b = mat.get_yn();
				// Which type of normalization? 
  if(norm_mode == 1) // L1_norm_mode.elem() == 1.0)
    {				// Loop through each row. 
      for(j = 0; j < a; ++j)
	{
	  count = 0.0;		// Sum elements in row j.
	  for(i = 0; i < b; ++i)
	    count += fabs(mat.elem(j,i));

				// If too small, then don't normalize 
	  if(count > 0.0001)
				// Apply normalization factor.
	    for(i = 0; i < b; ++i)
	      mat.elem(j,i) = mat.elem(j,i)/count;
	}
    }
  else				// L2-norm. 
    {				// Loop through each row.
      for(j = 0; j < a; ++j)
	{
	  count = 0.0;
				// Sum elements in the row.
	  for(i = 0; i < b; ++i)
	    count += Square(mat.elem(j,i));
	  count = sqrt(count);
				// If too small, then don't normalize.
	  if(count > 0.0001)
				// Apply normalization factor.
	    for(i = 0; i < b; ++i)
	      mat.elem(j,i) = mat.elem(j,i)/count;
	}
    }
 } 


 void normal_col(NslMatrix& mat, int norm_mode)
 {
  int i,j;
  int a,b;
  float count;
				// Get size of matrix
  a = mat.get_xn();
  b = mat.get_yn();
				// Check type of normalization.
  if(norm_mode == 1) // L1_norm_mode.elem() == 1.0)
    {				// Loop through each column.
      for(i = 0; i < b; ++i)
	{
	  count = 0.0;		// Sum elements in column i.
	  for(j = 0; j < a; ++j)
	    count += fabs(mat.elem(j,i));

				// If too small, then don't normalize
	  if(count > 0.0001)
				// Apply normalization factor.
	    for(j = 0; j < a; ++j)
	      mat.elem(j,i) = mat.elem(j,i)/count;
	}
    }
  else				// L2-norm.
    {				// Loop through each column.
      for(i = 0; i < b; ++i)
	{
	  count = 0.0;		// Sum elements in column i.
	  for(j = 0; j < a; ++j)
	    count += Square(mat.elem(j,i));
	  count = sqrt(count);

				// If too small, then don't normalize
	  if(count > 0.0001)
				// Apply normalization factor.
	    for(j = 0; j < a; ++j)
	      mat.elem(j,i) = mat.elem(j,i)/count;
	}
    }
 }


 void select_random(NslVector& vec, float prob)
 {
	int i,r;
	int init;
	int limit;

	init = vec.get_x0();
	limit = vec.get_x1();
	for(i = init; i <= limit; ++i) 
	{
	 	r = (random_value2() < prob);
		vec.elem(i) = r;
	}
 }

 

 void select_random(NslMatrix& mat, float prob)
 {
	int i, j;
	float s,r;
	int x0, x1;
	int y0, y1;

	x0 = mat.get_x0();
	x1 = mat.get_x1();
	y0 = mat.get_y0();
	y1 = mat.get_y1();
				
	for(i = x0; i <= x1; ++i)
		for(j = y0; j <= y1; ++j) 
		{
			r = random_value2();
			s = (r < prob);
			mat.elem(i,j) = s;
		}
 }
}


nslModule Feature() extends NslModule()
{
	public NslDinFloat1 inputs();
	public NslDinFloat1 voting_contribution();
	public NslDinFloat0 threshold_v();
	public NslDinInt0 factor();
	
	public NslDoutFloat1 feature();	 	
	
	private NslFloat1 feature_mem(); 		
	private NslFloat0 threshold_f();		
	private NslFloat0 u_feature();		 
	private NslFloat0 lrate_f(); 
	
	private NslFloat2 w_in_feature();
	private NslFloat2 dw_in_feature();
	private NslFloat2 w_in_feature_mask();
	private NslFloat0 input_weight_bias();	
	private NslFloat0 w_in_featureProbability();	
	private NslFloat0 L1_norm_mode();		
	private NslFloat0 negative_factor_f();	

 public void memAlloc(int num_inputs,int num_columns)
 {
	inputs.memAlloc(num_inputs);
	voting_contribution.memAlloc(num_columns);
	feature.memAlloc(num_columns);
	feature_mem.memAlloc(num_columns);
	w_in_feature.memAlloc(num_inputs,num_columns);
	dw_in_feature.memAlloc(num_inputs,num_columns);
	w_in_feature_mask.memAlloc(num_inputs,num_columns);
 }

 public void initModule()
 {
  	randomize(w_in_feature);
	LearnMethods.select_random(w_in_feature_mask,w_in_featureProbability.elem());
	w_in_feature = (w_in_feature/2.0 + 0.5 + input_weight_bias) 
		^ w_in_feature_mask;
	LearnMethods.normal_col(w_in_feature, L1_norm_mode.elem());
 }

 public void initTrain()
 {
  	feature_mem = -threshold_v;
  	feature = 0;
 }

 public void simTrain()
 {	
  	nslDiff(feature_mem, u_feature, -feature_mem - threshold_f +
       	mat_mult_col_vec(w_in_feature, inputs));
       
	if (LIMITED_ACTIVITY_FLAG == true)
  		feature = NSLsat(feature_mem, 0, 1, 0, 1);
	else
  		feature = NSLramp(feature_mem);
 }


 public void endTrain()
 {   	
  	float f_factor;

	if(factor.elem() < 0.0)
    		f_factor = factor.elem() * negative_factor_f.elem();
  	else
    		f_factor = factor.elem();

  	dw_in_feature = f_factor * lrate_f * 				
		(vec_mult_vec_trans(voting_contribution, inputs) 
    		^ w_in_feature_mask);

	// Keep weights within [0,1] range
  	w_in_feature = NSLramp(w_in_feature + dw_in_feature);

	// Normalize weights
  	LearnMethods.normal_col(w_in_feature,L1_norm_mode.elem());
 }
}



nslModule Noise() extends NslModule()
{
	public Noise noise();
	public NslFloat0 gain();
	public NslFloat0 change_probability();

 public void memAlloc(int num)
 {
	noise.memAlloc(num);
 }

 public void initExec()
 {
  	randomize(noise);		// Intialize noise vectors
  	noise = gain * noise;
 }
 public void simExec()
 {
	// Roll die to see if change is appropriate
  	if(nslRand() < change_probability.elem())
    	{
      	randomize(noise);
      	noise = gain * noise;
    	}
 }
}

 

nslModule Threshold() extends NslModule()
{
	public NslDinFloat0 s();
	public NslDoutFloat0 threshold_v();	// Threshold (varying over time) 

	private NslFloat0 u_threshold_v();// Time constant of threshold change. 
	private NslFloat0 init_threshold_v();

 public void initTrain()
 {
  	threshold_v = init_threshold_v;
 }
 public void simTrain()
 {
  	if(s.elem() <= 0.0)
    		nslDiff(threshold_v,u_threshold_v, -threshold_v); 
 }
}

 

nslModule Vote() extends NslModule()
{
	public Noise noise();
	
	public NslDinFloat0 threshold_v();
	public NslDinFloat1 feature();
	public NslDoutFloat1 voting();
	public NslDoutFloat1 voting_contribution(); // Output of voting units. 
			
	private NslFloat1 voting_mem();
	private NslFloat0 u_voting();

	private NslFloat0 voting_contributionmode();	
	private NslFloat0 votingContributionScale(); 
	private NslFloat0 display_participation_mode(); 
	private NslFloat1 voting_participation(); 	                                      	
	private NslFloat0 init_threshold_v();
	
	private int _num_columns;

 public void memAlloc(int num_columns)
 {
	_num_columns = num_columns;
	
	noise.memAlloc(num_columns);
	feature.memAlloc(num_columns);
	voting.memAlloc(num_columns);
	voting_contribution.memAlloc(num_columns);
	voting_participation.memAlloc(num_columns);
	voting_mem.memAlloc(num_columns);
 }

 public void initModule()
 {
  	noise.initExec();
 }
 public void initTrain()
 {
  	voting_mem = -init_threshold_v;
  	voting = 0;
 }

 public void simTrain()
 {
 	noise.simExec();
	
  	nslDiff(voting_mem, u_voting, -voting_mem - threshold_v + 
  		feature + noise.noise);

	if (LIMITED_VOTING_ACTIVITY == true) noise.simExec();
		voting = NSLsat(voting_mem, 0, 1, 0, 1);
	else
  		voting = NSLramp(voting_mem);
		
	voting_participation = NSLstep(voting);


	if (voting_contributionmode.elem() == LINEAR) 
    		voting_contribution = voting; 
  	else if (voting_contributionmode.elem() == BINARY)
		voting_contribution = voting_participation;
    	else if (voting_contributionmode.elem() == COMPRESSED_LINEAR)
	    	voting_contribution = NSLsat(voting,0.0,votingContributionScale.elem(),0.0, 1.0);
	else if (voting_contributionmode.elem() == JUMP_LINEAR)
		voting_contribution = NSLsat(NSLramp(voting, 0.0, 0.0,votingContributionScale.elem()));
	else
   		nslPrintln("Unknown voting_contribution_mode: ",votingContributionScale.elem());
 }

 public void endTrain()
 {
 	nslPrintln("Average Voting: ",voting.sum()/_num_columns);
 	  
  	if (display_participation_mode.elem() == 1.0) 
  		voting_participation.print(cout);
 }
}


nslModule Motor() extends NslModule()
{
	public Noise noise();
	
	public NslDinFloat1 voting();
	public NslDinFloat1 voting_contribution();
	public NslDinFloat0 s();
	public NslDinInt0 factor();
	
	public NslDoutFloat1 motor();

	public NslDoutInt0 above_thresh_num();    	
	public NslDoutInt0 stable_num();    	
	public NslDoutInt0 winner();    	
		
	private NslFloat1 dmotor_mem();
	private NslFloat1 motor_mem();
	private NslFloat1 motor_inputs();
	private NslFloat0 u_motor();
	private NslFloat0 threshold_m();
	
	private NslFloat2 w_vote_motor();
	private NslFloat2 dw_vote_motor();
	private NslFloat2 w_vote_motor_mask();
	private NslFloat0 voting_weight_bias();
	private NslFloat0 w_vote_motorProbability();
	private NslFloat0 voting_factor(); 
	private NslFloat0 lrate_v(); 
		
	private NslFloat0 L1_norm_mode();			
	private NslFloat0 normalize_input_mode();	
	private NslFloat0 stable_detect_threshold(); 
  	private int _num_columns;
	private int _num_choices;
 		                                      					
 public void memAlloc(int num_columns,int num_choices)
 {
	_num_columns = num_columns;
	_num_choices = num_choices;
	
	noise.memAlloc(num_choices);
	motor.memAlloc(num_choices);
	voting.memAlloc(num_columns);
	voting_contribution.memAlloc(num_columns);
	motor_mem.memAlloc(num_choices);
	dmotor_mem.memAlloc(num_choices);
	motor_inputs.memAlloc(num_choices);
	w_vote_motor.memAlloc(num_columns,num_choices);
	dw_vote_motor.memAlloc(num_columns,num_choices);
	w_vote_motor_mask.memAlloc(num_columns,num_choices);
 }

 public void initModule()
 {
  	noise.initExec();
    	winner = 0;

  	randomize(w_vote_motor);
  	LearnMethods.select_random(w_vote_motor_mask, w_vote_motorProbability.elem());
	// select_random(w_vote_motor_mask, w_vote_motorProbability.elem());
  	w_vote_motor = (w_vote_motor/2.0 + 0.5 + voting_weight_bias) ^ w_vote_motor_mask;
				
	// Normalize as appropriate
  	if (normalize_input_mode.elem() == 1.0)
    	LearnMethods.normal_col(w_vote_motor,L1_norm_mode.elem());
  	else
    	LearnMethods.normal_row(w_vote_motor,L1_norm_mode.elem());
 }

 public void initTrain()
 {
  	motor_mem = -threshold_m;
  	motor = 0;

	above_thresh_num = 0;
	stable_num = 0;
 }

 public void simTrain()
 {	
    	noise.simExec();
    
  	motor_inputs = mat_mult_col_vec(w_vote_motor, voting)/_num_columns;
  	dmotor_mem = -motor_mem - threshold_m +
    		voting_factor * motor_inputs - s + motor + noise.noise;
  	nslDiff(motor_mem, u_motor,dmotor_mem);
  	motor = NSLstep(motor_mem);
   		
	above_thresh_num = 0;
	stable_num = 0;
  
      // Check the status of each motor program selection unit. 
  	for (int i = 0; i < _num_choices; ++i)
    	{
      	if (motor.elem(i) > 0.0)	// Above threshold
		{
	  		above_thresh_num = above_thresh_num + 1;
	  		winner = i;
		}
		// No change in activity.  - used only in standard WTA.
		if(fabs(dmotor_mem.elem(i)) < stable_detect_threshold.elem())
			stable_num = stable_num + 1;
    	}
 }

 
 public void endTrain()
 {
  	dw_vote_motor = factor * lrate_v *
    	(vec_mult_vec_trans(motor, voting_contribution) ^ w_vote_motor_mask);

	// Keep weights within [0,1] range
  	w_vote_motor = NSLramp(w_vote_motor + dw_vote_motor);

	// Normalize weights, depending upon mode
  	if(normalize_input_mode.elem() == 1.0)
    	LearnMethods.normal_col(w_vote_motor,L1_norm_mode.elem());
  	else
    	LearnMethods.normal_row(w_vote_motor,L1_norm_mode.elem());
 }
}


nslModule WTA() extends NslModule()
{
	public NslDinFloat1 motor();
	public NslDoutFloat0 s();
		
	private NslFloat0 s_mem();
	private NslFloat0 u_s();
	private NslFloat0 threshold_s();

 public void memAlloc(int num_choices)
 {
	motor.memAlloc(num_choices);
 }

 public void initTrain()
 {
  	s_mem = 0;
  	s = 0;
 }

 public void simTrain()
 {
  	nslDiff(s_mem, u_s, - s_mem - threshold_s + motor.sum());
  	s = NSLramp(s_mem);
 }
}


nslModule CondLearn() extends NslModule()
{
  	public Feature feature();
  	public Threshold threshold();
  	public Vote vote();
  	public Motor motor();
  	public WTA wta();
  	// seguti
	public NslDinInt1 input1();    		
	public NslDinInt0 output1(); 
	//seguti 
	public NslDoutInt0 new_read();	//  if 1 read new input
	public NslDoutInt0 factor();
			  	
	public NslDinInt0 above_thresh_num();    	
	public NslDinInt0 stable_num();    	
	public NslDinInt0 winner();    	
			
	private NslInt0 collect();
	private NslFloat0 display_set_unit_flag(); 
	private NslFloat0 first_pole_mode();	
	private NslFloat0 normalize_input_mode();
	private NslFloat0 special_punish_mode(); 
	private NslInt0 seed();
	
	int _num_columns;
	int _num_choices;
	private int collect_mode; 
	private int timer_flag;	 	

 public void memAlloc(int inSize, int num_columns, int num_choices)
 {
	_num_columns = num_columns;
	_num_choices = num_choices;
	
	input1.memAlloc(inSize);
	feature.memAlloc(inSize,num_columns);
	vote.memAlloc(num_columns);
	motor.memAlloc(num_columns,num_choices);
	wta.memAlloc(num_choices);
 }

 public void makeConn() 
 {
	nslConnect(feature.feature,vote.feature);
	nslConnect(vote.voting,motor.voting);
	nslConnect(threshold.threshold_v,vote.threshold_v);
	nslConnect(motor.motor,wta.motor);
	nslConnect(wta.s,motor.s);
	nslConnect(vote.voting_contribution,motor.voting_contribution);
	nslConnect(vote.voting_contribution,feature.voting_contribution);

	nslConnect(factor,feature.factor);
	nslConnect(factor,motor.factor);	
	
	nslRelabel(input1,feature.inputs); 

	nslRelabel(motor.stable_num,stable_num); 
	nslRelabel(motor.above_thresh_num,above_thresh_num); 
	nslRelabel(motor.winner,winner); 
	
	int s = seed.get_value(); // long
	
	nslRand(s);
 }

 public void initTrain()
 {
	collect_mode = collect.elem();
  	if(!collect_mode)
    	cmd_out("Cleared");
    	   
    	timer_flag = 0;
 }


 public void simTrain()
 {
	collect_mode = collect.elem();
	
	// Exactly one unit active and in first pole mode or all units are stable. 
  	if(above_thresh_num.elem() == 1 &&
     		(stable_num.elem() == _num_choices || first_pole_mode.elem() != 0.0))
    {
    	timer_flag = 1;
      	if(!collect_mode)
			cmd_out("Done!!");
		if(collect_mode) {	// Update weights and set up next trial 
			punish_reward_func(winner.elem()); 
			nslBreakSim();
		}
    }
 }

 public void endTrain()	// Timeout: NO-GO case
 {
	if (timer_flag == 1)
		return;
		
	// On completion, set to 1 if at least  1 MPSU is active, -1 if none

	collect_mode = collect.elem();
					
    if(above_thresh_num.elem() == 0)
		timer_flag = -1;	// Signal no activity at MPSU level. 
    else
      	timer_flag = 1;

    if(!collect_mode) {
		nslPrintln("No-pick no-go!");
		if (timer_flag == 1)
			nslPrintln("Multiple picked!");
	}
    else {	// Update weights and set up next trial 
		punish_reward_func(NO_GO_CASE);
		nslBreakSim();
	}
 }


 public void punish_reward_func(int win)
 {

	// If no MPSUs are active, then modify the reward signal so it is positive. 
	// This is done so that some column will be more likely to be active next
	// time th esame input is give.
	
	factor = 1;
	
  	nslPrintln("Time: ",NSL_SYSTEM->getTrainTimeStep());
  	nslPrintln("Output: ",output1.elem());
    	nslPrintln("Winner: ",win);

  	if(output1.elem() == win)
    	{
 		new_read = 1;
  		if(!collect_mode)
    		nslPrintln("Rewarded");
      	nslPrintln("+");
    	}
  	else
    	{
    		new_read = 0;
    		if (timer_flag != -1)
 			factor = -1;
  		if(!collect_mode)
    		nslPrintln("Punished");
      	nslPrintln("-");    
   	}  
 }
}


nslModule TrainFile() extends NslModule()
{
	public NslDinInt0 new_read();	//  if 1 read new input
	public NslDoutInt1 input1();    	//  Current vector of inputs
	public NslDoutInt0 output1();    	//  Current output
             		
	private NslFloat2 pInput();
	private int current_pat();            
                               		
	private NslFloat1 pOutput();	//  Set of expected outputs.  
	private int patterns_loaded_flag();
	
	private NslInt0 inSize();
	private NslInt0 outSize();
	private NslInt0 numPats();

	private NslInt0 repeat();
	private int repeat_mode(); 


 public void readFile()
 {
  	int i,iSize,oSize,nPats;
  	int val;
  
 	NslFile file("a1.pat");
				
  	if (file.open_file(NSL_INPUT) == -1) {
  		nslPrintln("Bad file name");
  		file.close_file();
  		return;
  	}
  	else {
  		file.read(nPats);
  		cmd_out("num_pats = ",nPats);
  		numPats = nPats;

  		file.read(oSize);
  		outSize = oSize;
  		if(oSize != 1)
    		{
      		nslPrintln("Wrong number of output units.");
      		nPats = 0;
      		file.close_file();
      		return;
    		}
  		file.read(iSize);		// Check number of input units. 
  		inSize = iSize;
  		if(iSize != inSize)
    		{
      		nslPrintln("Wrong number of input units.");
      		nslPrintln(iSize);
      		nPats = 0;
      		file.close_file();
      		return;
    		}
    		
    		input1.memAlloc(iSize);
		pInput.memAlloc(nPats,iSize);
		pOutput.memAlloc(nPats);
		
  		for(int pat = 0; pat < nPats; ++pat)
    		{
      		for(i = 0; i < inSize; ++i) 
      		{
	  			file.read(val);
	  			pInput.elem(pat, i) = val;
	  		}	
      		file.read(val);
      		pOutput.elem(pat) = val;
    		}
  		nslPrintln("patterns read: ", nPats);
  		patterns_loaded_flag = 1;
  		file.close_file();
   	}
 }

 

 public void initTrain()		
 {	
  	if(!patterns_loaded_flag)
    	{
      	nslPrintln("Must load patterns before simulation initialization and execution");
      	exit(0);
    	}

	if (NSL_SYSTEM->getEpochStep() == 0)
		repeat_mode = 0;
	else
		repeat_mode = repeat.elem();

	if (repeat_mode == 0 || new_read.elem() == 1)  
	{
    		int r = rand(); // rand()/100
    		int mod = (int) numPats.elem();
      	current_pat = r % mod;
      	extract_row_vector(pInput, current_pat, input1);
      }
 
 	//nslPrintln("New_Read: ",new_read.elem());  
	//nslPrintln("Current Pattern: ",current_pat);

	output1 = pOutput[current_pat];
 }
}


nslModule CondLearnModel() extends NslModel()
{
 	public TrainFile tf();
 	public CondLearn cl();


 public void makeConn()
 {
	nslConnect(tf.input1,cl.input1);
	nslConnect(tf.output1,cl.output1);
	nslConnect(cl.new_read,tf.new_read);
 }

 public void initModule()
 {
	tf.readFile();
		
	//seguti NslInt0 inSizeB = ((NslInt0&) tf.getDataRef("inSize"));
	int inSize = inSizeB.getValue();

//	int inSize = tf.getValue("inSize");
	int num_columns = 30;
	int num_choices = 4;	
			
	cl.memAlloc(inSize,num_columns,num_choices);
 }
}

