// ***************************************************************************
//			    NSL RETINA MODEL .C
//
//	Retina model description:
//	------------------------ 
//	Quantitative modeling of Responses of Anuran retina: Stimulus Shape and 
//	Size Dependency. In this simulations we test the model's ability to
//	reproduce quantitatively tabulated data on the dependency on stimulus
//	shape and size (Ewert 1976).
//	Input to the model is Light on the receptors (40X40 to simulate the 
//	receptieve field of a single ganglion cell of each type).
//	The model also simulates "simple" Horizontal and Bipolar cells.
// 	The output of the model the temporal firing rate of a  Ganglion cell of 
//    each type (R2, R3, R4). So the graphs given by NSL show the firing rate 
//	of a ganglion cell vs. the time.
//    
//	Notes:
//	-----
// 	No Trailing edge effect since Ewert computed his data with only
//	the response to the leading edge.
// 	No Leaky Integrators for the Ganglion Cells.
//
// 	Goal: Match Ewert's quantitative data on the Toad's retinal ganglion 
// 	cells.
//  	Status: goal achieved, paper submitted to Vision Research.
//
// 	File : "retina.c"
//      Contents: Definition of the neural net and main code. (see "retina.nsl")
// 	Last update: April 1st, 1992 
// 	By: Fernando Corbacho 
//
// ***************************************************************************
//#include "nsl_include.h"
//#include "retina.h"

// Modifications :
//	
//	1) Create nslClass RetinaLib
//	2) diff_ae(..) -> RetinaLib.diff_ae(..)
//	3) newconv(..) -> RetinaLib.newconv(..)
//	4) nslGauss(..) -> RetinaLib.nslGauss(..)
//	5) nsl_data -> NslData
//      6) nsl_matrix -> NslMatrix

import nsl_include;
import retina;


nslClass RetinaLib() extends NslObject()
{
  // AVERAGE EXACT METHOD 
 NslMatrix diff_ae(NslMatrix& V,NslMatrix& S,NslMatrix& Prev, NslData& tm)
 {
      float dt0,tc,tmp;
      int Vmax = V.get_imax();    // Size of the matrices.
         
	NslMatrix term1 (Vmax,Vmax);
	NslMatrix term2 (Vmax,Vmax);

	if (NSL_SYSTEM != NULL)
		dt0 = NSL_SYSTEM->getSimDelta();
	else
	{
		cmd_error("*** AVERAGE EXACT METHOD ***");
		cmd_error("Unknown Model");
		return term1;
	}

	if (dt0 == 0 || tm.get_data() == 0)
	{
		cmd_error("*** AVERAGE EXACT METHOD ***");
		cmd_error("delta: ",dt0);
		cmd_error("time constant: ",tm.get_data());
		cmd_error("Bad Integration Parameters");
		return term1;
	}

        // **************** This is the real Calculation *************** 

	tc = tm.get_data(); // time constant.
	tmp = exp(-dt0/tc);

	term1 = (NslMatrix) ( (1 - tmp) * S + tmp * V ) ;
	term2 = (NslMatrix) ( (S - Prev) * (tmp * (tc + dt0) - tc) / dt0); 
	 
	return (NslMatrix) (term1 + term2);
 }

 // a is the Mask and b is the input layer.
 NslMatrix newconv(NslMatrix& a,NslMatrix& b)
 {
      int saimax = a.get_imax();
      int sajmax = a.get_jmax();
      int sbimax = b.get_imax();
      int sbjmax = b.get_jmax();

      int leftbound = 1;       // 32; for the 72x72
      NslMatrix c(1,1); // Make this variable size.
			   // c(8,8) for 72x72

      for (int i = 0; i < leftbound; i = i+4)
	{
           for (int j = 0; j < leftbound; j = j+4)
           {    
                num_type val = (num_type) 0;

                for (int m = 0; m < saimax; m++)
                   for (int n = 0; n < sajmax; n++)
                        val = val + a.elem(m,n) * b.elem(i+m,j+n);
                c.elem(i/4,j/4) = val;
			
           }
	}
      return c;
 }

 NslMatrix nslGauss(int imax,int jmax,NslData& dia,NslData& sig,NslData& wgt)
 {
        int i,j;
        num_type rr,rad;

        NslMatrix Q(imax,jmax);

	//int imax = Q.get_imax();
	//int jmax = Q.get_jmax();
        int imax2 = imax/2;
        int jmax2 = jmax/2;
             
        num_type rad0 = dia.elem()/2.0;
        num_type sum = 0.0;
        if (dia.elem() == 0.0)
                return Q;

        for (i=0; i<imax; i++)
            for (j=0; j<jmax; j++)
            {
                // rr is in units of 1/2 degrees, same as kernel array 
                rr = (num_type) sqrt((num_type) pow((num_type)(i-imax2),2.0) + 
                        (num_type) pow((num_type)(j-jmax2),2.0));
                // rad is in units of degrees 
                rad = rr/2.0;
                if (rad < rad0)
                     Q.elem(i,j) = (num_type) 
                        exp(-(rad*rad)/(2.0*sig.elem()*sig.elem()));
                sum = sum + Q.elem(i,j);
            }

        if (sum == 0.0)
                Q = 0;
        else 
                Q = Q/sum*wgt;

        return Q;
 }
}

extern nslModule NslStimulusFloat2;

nslModule Visin(int size1,int size2) extends NslModule()
{
	public NslDoutFloat2 out(size1,size2);
	private NslStimulusFloat2 in(size1,size2);

 public void initRun()
 {
	in = 0;
 }

 // stimulus(r, r0, c0, r1,c1, vx ,vy, flag); // if needed blurred edge effect.
 
 public void simRun()	
 {
	in.run();	//Get the stimuli. Defined in the file "retina.nsl"
	out = in;
 }
}


nslModule Receptor(int size1,int size2) extends NslModule()
{
	public NslDinFloat2 in(size1,size2); 	
	public NslDoutFloat2 r(size1,size2);	// Receptors
	private NslFloat2 rB(size1,size2);	// For sustained response.

 public void initRun()	// **** Receptors ****
 {
	r = 0;
 }

 public void simRun()	// **** Receptors ****
 {
	float t = NSL_SYSTEM->getSimTime();
	
	rB = r;
	// stimulus(r, r0, c0, r1,c1, vx ,vy, flag); // if needed blurred edge effect.
 
	if (t <= 7.0)	      //1.55 in order to stop it in the ERF.
   	{
		r = in;	//Get the stimuli. Defined in the file "retina.nsl"
   	}
  	else
		r = rB;
 }
}


nslModule Horizontal(int size1,int size2) extends NslModule()
{
	public NslDoutFloat2 hor(size1,size2) ;	// Horizontal cells
	private NslFloat0 hor_tm();	// Horizontal time constant
	private NslFloat0 H_level(); // Uniform horizontal potential
                        //   (0 if the background is bright, 1 if dark.)

 public void initRun()	// **** HORIZONTAL CELLS ****
 {
	hor = 0;
 }

 public void simRun()	// **** HORIZONTAL CELLS ****
 {
	nslDiff(hor,hor_tm,H_level - hor);
 }
}


nslModule BipolarOff(int size1,int size2) extends NslModule()
{
	public NslDinFloat2 r(size1,size2) ;	// Receptors
	public NslDinFloat2 hor(size1,size2);// Horizontal cells
	public NslDoutFloat2 hbc(size1,size2);
	public NslDoutFloat2 pbh(size1,size2);

 public void simRun()	// **** HYPERPOLARIZING BIPOLARS ****
 {
	hbc = r - hor;
	pbh = NSLmax(hbc,0);		
 }
}


nslModule BipolarOn(int size1,int size2) extends NslModule()
{
	public NslDinFloat2 r(size1,size2) ;	// Receptors
	public NslDinFloat2 hor(size1,size2) ;	
	public NslDoutFloat2 dbc(size1,size2) ;	
	private NslFloat2 pbd(size1,size2) ;

 public void simRun()	// **** DEPOLARIZING BIPOLARS ****
 {
	dbc = hor - r;
	pbd = NSLmax(dbc,0);
 }
}


nslModule AmacrineOff(int size1,int size2) extends NslModule()
{
	public NslDinFloat2 hbc(size1,size2) ;	
	private NslFloat2 prev(size1,size2) ; 
	private NslFloat2 axh(size1,size2) ; 	
	private NslFloat2 ath1(size1,size2) ; 
	private NslFloat2 ath2(size1,size2) ; 
	public NslDoutFloat2 ath(size1,size2) ;
	public NslDoutFloat2 old_ath(size1,size2) ;	
	private NslFloat0 axh_tm() ;		

 public void initRun()	// **** Off-Channel Amacrine Cells ****
 {
	prev = 0;
	axh = 0; ath1 = 0; ath2 = 0; ath = 0; 
	dt = NSL_SYSTEM->getSimDelta();
 }
 public void simRun()	// **** Off-Channel Amacrine Cells ****
 {
	// DIFF.eq(axh,axh_tm) = hbc - axh; 	// Euler method.
        axh = RetinaLib.diff_ae(axh,hbc,prev,axh_tm);	// Average Exact Method.	
	prev = hbc;

	old_ath = ath;
	ath1 = 5*(hbc - axh);
	ath2 = exp(-dt/axh_tm.elem())*old_ath;

	ath = NSLmax(ath1,ath2) ;
 }
}


nslModule AmacrineOn(int size1,int size2) extends NslModule()
{
	public NslDinFloat2 dbc(size1,size2) ;	
	private NslFloat2 prev(size1,size2) ; 
	private NslFloat2 axd(size1,size2) ; 	
	private NslFloat2 atd1(size1,size2) ; 
	private NslFloat2 atd2(size1,size2) ;
	public NslDoutFloat2 atd(size1,size2) ; 
	public NslDoutFloat2 old_atd(size1,size2) ; 
	private NslFloat0 axd_tm() ;

 public void initRun()	// **** On-Channel Amacrine Cells ****
 {
	prev = 0;
 	axd = 0; atd1 = 0; atd2 = 0; atd = 0;
	dt = NSL_SYSTEM->getSimDelta();
 }

 public int simRun()	// **** On-Channel Amacrine Cells ****
 {
	// DIFF.eq(axd,axd_tm) = dbc - axd;
	axd = RetinaLib.diff_ae(axd, dbc, prev, axd_tm); 
	prev = dbc;

	old_atd = atd;
	atd1 = 5*(dbc - axd);             // This is because no BIPOLARS.
 	atd2 = exp(-dt/axd_tm.elem())*old_atd;

	atd = NSLmax(atd1,atd2);  
 } 
}


nslModule R2(int size1,int size2) extends NslModule()
{
	public NslDinFloat2 pbh(size1,size2) ;
	public NslDinFloat2 ath(size1,size2) ;
	public NslDinFloat2 atd(size1,size2) ; 
	private NslFloat2 r2(1,1) ; 
	private NslFloat2 r2_i(1,1) ;	
	private NslFloat2 sust(1,1) ; 
	private NslFloat2 sust_irf(1,1) ;
	private NslFloat2 sust_erf(1,1) ;
	private NslFloat2 r2t(1,1) ;
	private NslFloat2 r2f(1,1) ;	// Scaled R2.
	private NslFloat2 pbh_irf() ;
	private NslFloat2 pbh_erf() ;
	private NslFloat2 erf(size1,size2 ) ; 
	private NslFloat2 irf(size1,size2 ) ; 
	private NslFloat2 rf(size1,size2 ) ;		
	private NslFloat0 erf_dia() ;		
	private NslFloat0 irf_dia() ;
	private NslFloat0 erf_sig() ;		
	private NslFloat0 irf_sig() ;
	private NslFloat0 erf_wgt() ;		// Weight.
	private NslFloat0 irf_wgt() ;
	private NslFloat0 thresh() ;		
	private NslFloat0 k() ;			
	private NslFloat0 tm() ;		
	private NslFloat0 trailing() ;
	private NslFloat2 temp(size1,size2) ;
	private NslFloat2 stat(3,3) ;  		

 public void initRun()
 {
	r2 = 0; r2t = 0; r2f = 0; 
	
	int size1 = erf.get_imax();
	int size2 = erf.get_jmax();
	
 	// Construct the Gaussian kernels for the ganglion cells.
	erf = RetinaLib.nslGauss(size1,size2,erf_dia,erf_sig,erf_wgt);
	irf = RetinaLib.nslGauss(size1,size2,irf_dia,irf_sig,irf_wgt);

	rf = erf - irf;            // DOG for the r2 ganglion cells
 }
 public void simRun()
 {
	temp = ath + trailing * atd;   // trailing is the effect of the trailing
				       // edge set to 0 to get Ewert's data.

	// sustained input.
        sust_erf = RetinaLib.newconv(erf, pbh_erf * pbh);
	sust_irf = RetinaLib.newconv(irf, pbh_irf * pbh);

	sust = sust_erf - sust_irf;	

	r2 = RetinaLib.newconv(rf, temp) + sust;    // New convolution and No Leaky Integ.
	
	r2_i = r2;

	r2t = NSLramp(r2);
	r2f = k*r2t;		// Scaling.

	if  (r2f.elem(0,0) >  thresh.elem())  //  ****** STATISTICS ******
	{
		// Compute the average firing rate.
		stat.elem(0,0) = stat.elem(0,0) + r2f.elem(0,0);
		stat.elem(0,1)++;
		stat.elem(0,2) = stat.elem(0,0) / stat.elem(0,1);
	}
 }
}


nslModule R3(int size1,int size2) extends NslModule()
{
	public NslDinFloat2 ath(size1,size2) ;
	public NslDinFloat2 atd(size1,size2) ; 
	public NslDinFloat2 old_ath(size1,size2) ;
	public NslDinFloat2 old_atd(size1,size2) ; 
	private NslFloat2 r3(1,1) ; 
	private NslFloat2 r3t(1,1) ;
	private NslFloat2 r3f(1,1) ; 
	private NslFloat2 erf(size1,size2 ) ;
	private NslFloat2 irf(size1,size2 ) ;
	private NslFloat0 erf_dia() ;
	private NslFloat0 irf_dia() ;
	private NslFloat0 erf_sig() ;
	private NslFloat0 irf_sig() ;
	private NslFloat0 erf_wgt() ;
	private NslFloat0 irf_wgt() ;
	private NslFloat0 thresh() ;
	private NslFloat0 k() ;			// outputs.
	private NslFloat0 tm() ;
	private NslFloat0 p() ;	
	private NslFloat2 all(size1,size2) ;	
	private NslFloat2 old(size1,size2) ;	
	private NslFloat2 stat(3,3) ;  	

 public void initRun()
 {
	r3 = 0; r3t = 0; r3f = 0;

	int size1 = erf.get_imax();
	int size2 = erf.get_jmax();
	
	erf = RetinaLib.nslGauss(size1,size2,erf_dia,erf_sig,erf_wgt);
	irf = RetinaLib.nslGauss(size1,size2,irf_dia,irf_sig,irf_wgt);
 }
 public void simRun()
 {
	all = p * atd + ath;
	old = p * old_atd + old_ath;
	
        // DIFF.eq(r3,tm) = -r3 + RetinaLib.newconv(erf, all) - RetinaLib.newconv(irf, old); 
	r3 = RetinaLib.newconv(erf, all) - RetinaLib.newconv(irf, old);

	r3t = NSLramp(r3);
	r3f = k*r3t;

	if (r3f.elem(0,0) > thresh.elem())   //  ***** STATISTICS *****
	{
		stat.elem(1,0) = stat.elem(1,0) + r3f.elem(0,0);
		//   stores the summation of the responses.
		stat.elem(1,1)++;
		//   stores the number of time steps .
		stat.elem(1,2) = stat.elem(1,0) / stat.elem(1,1);
		//   stores the Average Firing Rate up to the moment.
	}
 }
}


nslModule R4(int size1,int size2) extends NslModule()
{
	public NslDinFloat2 ath(size1,size2) ;
	public NslDinFloat2 atd(size1,size2) ; 
	private NslFloat2 r4(1,1) ; 
	private NslFloat2 r4t(1,1) ; 
	private NslFloat2 r4f(1,1) ;	// Scaled R4.
	private NslFloat0 erf_dia() ;
	private NslFloat0 erf_sig() ;
	private NslFloat0 erf_wgt() ;
	private NslFloat2 erf(size1,size2) ; 
	private NslFloat0 thresh() ;
	private NslFloat0 k() ;
	private NslFloat0 tm() ;
	private NslFloat2 stat(3,3) ;  		

 public void initRun()
 {
	r4 = 0; r4t = 0; r4f = 0;

	int size1 = erf.get_imax();
	int size2 = erf.get_jmax();
	
	// Construct the Gaussian kernels for the ganglion cells.
	erf = RetinaLib.nslGauss(size1,size2,erf_dia,erf_sig,erf_wgt);
 }

 public void simRun()
 {
	// DIFF.eq(r4,r4_tm) = -r4 + RetinaLib.newconv(erf,(ath - atd)); 
        r4 = RetinaLib.newconv(erf, (ath - atd)); 

	r4t = NSLramp(r4);
	r4f = k*r4t/(r4t+0.2);  // Squashing function.
  
	if (r4f.elem(0,0) > thresh.elem())   //  ***** STATISTICS *****
	{
		stat.elem(2,0) = stat.elem(2,0) + r4f.elem(0,0);
		stat.elem(2,1)++;
		stat.elem(2,2) = stat.elem(2,0) / stat.elem(2,1);
	}
 }
}


nslModule Retina(int size1,int size2) extends NslModule()
{
	public Visin visin(size1,size2) ;
	public Receptor receptor(size1,size2) ;
	public  Horizontal horizontal(size1,size2) ;
	public BipolarOn bipolarOn(size1,size2) ;
	public BipolarOff bipolarOff(size1,size2) ;
	public AmacrineOn amacrineOn(size1,size2) ;
	public AmacrineOff amacrineOff(size1,size2) ;
	public R2 r2(size1,size2);	
	public R3 r3(size1,size2);	
	public R4 r4(size1,size2);

 public void makeConn()
 {
	nslConnect(visin.out,receptor.in);
	nslConnect(receptor.r,bipolarOn.r);
	nslConnect(horizontal.hor,bipolarOn.hor);
	nslConnect(receptor.r,bipolarOff.r);
	nslConnect(horizontal.hor,bipolarOff.hor);
	nslConnect(bipolarOn.dbc,amacrineOn.dbc);
	nslConnect(bipolarOff.hbc,amacrineOff.hbc);
	nslConnect(bipolarOff.pbh,r2.pbh);
	nslConnect(amacrineOff.ath,r2.ath);
	nslConnect(amacrineOn.atd,r2.atd);
	nslConnect(amacrineOff.ath,r3.ath);
	nslConnect(amacrineOn.atd,r3.atd);
	nslConnect(amacrineOff.old_ath,r3.old_ath);
	nslConnect(amacrineOn.old_atd,r3.old_atd);
	nslConnect(amacrineOff.ath,r4.ath);
	nslConnect(amacrineOn.atd,r4.atd);
 }
}


nslModel RetinaModel extends NslModel ()
{
	public Retina retina(40,40);
}

